declare namespace firebase {

    /**
     * The allowed values for LoginOptions.type.
     */
    export enum LoginType {
        /**
         * No further data is required.
         */
        ANONYMOUS,
            /**
             * This requires you to pass in email and password properties as well.
             */
        PASSWORD,
            /**
             * This requires you to pass either an authentication token generated by your backend server
             * or the tokenProviderFn function that returns a promise to provide the token.
             * See: https://firebase.google.com/docs/auth/server
             */
        CUSTOM,
            /**
             * This requires you to setup Facebook Auth in the Firebase console,
             * as well as uncommenting the SDK includes in include.gradle (Android) and Podfile (iOS).
             */
        FACEBOOK,
            /**
             * This requires you to setup Google Sign In in the Firebase console,
             * as well as uncommenting the SDK includes in include.gradle (Android) and Podfile (iOS).
             */
        GOOGLE
    }

    /**
     * The allowed values for QueryOptions.orderBy.type.
     */
    export enum QueryOrderByType {
        KEY,
        VALUE,
        CHILD,
        PRIORITY
    }

    /**
     * The allowed values for QueryOptions.range.type.
     */
    export enum QueryRangeType {
        START_AT,
        END_AT,
        EQUAL_TO
    }

    /**
     * The allowed values for QueryOptions.limit.type.
     */
    export enum QueryLimitType {
        FIRST,
        LAST
    }

    export interface ServerValue {
        /**
         * When for instance using setValue you can set a timestamp property to this placeholder value.
         * Example:
         *   updateTs: firebase.ServerValue.TIMESTAMP
         */
        TIMESTAMP: any;
    }

    /**
     * The options object passed into the init function.
     */
    export interface InitOptions {
        /**
         * Allow disk persistence. Default false.
         */
        persist?: boolean;
        /**
         * Get notified when the user is logged in.
         */
        onAuthStateChanged?: (data: AuthStateData) => void;
        /**
         * Attempt to sign out before initializing, useful in case previous
         * project token is cached which leads to following type of error:
         *   "[FirebaseDatabase] Authentication failed: invalid_token ..."
         * Default false.
         */
        iOSEmulatorFlush?: boolean;
        /**
         * For Firebase Storage you can pass in something like 'gs://n-plugin-test.appspot.com'
         * here so we can cache it. Otherwise pass in the 'bucket' param when using Storage features.
         * Can be found in the firebase console.
         */
        storageBucket?: string;
    }

    export interface QueryRangeOption {
        type: QueryRangeType;
        value: string;
    }

    /**
     * The options object passed into the query function.
     */
    export interface QueryOptions {
        /**
         * How you'd like to sort the query result.
         */
        orderBy: {
            type: QueryOrderByType;
            /**
             * mandatory when type is QueryOrderByType.CHILD
             */
            value?: string;
        };

        /**
         * You can further restrict the returned results by specifying restrictions.
         * Need more than one range restriction? Use 'ranges' instead.
         */
        range?: QueryRangeOption;

        /**
         * Same as 'range', but for a 'chain of ranges'.
         * You can further restrict the returned results by specifying restrictions.
         */
        ranges?: QueryRangeOption[];

        /**
         * You can limit the number of returned rows if you want to.
         */
        limit?: {
            type: QueryLimitType;
            value: number;
        };

        /**
         * Set this to true if you don't want to listen for any future updates,
         * but just want to retrieve the current value.
         * You can also use this to check if certain data is in the database.
         * Default false.
         */
        singleEvent?: boolean;
    }

    export interface GetAuthTokenOptions {
        /**
         * Default false.
         */
        forceRefresh?: boolean;
    }

    /**
     * The options object passed into the login function.
     */
    export interface LoginOptions {
        type: LoginType;
        email?: string;
        password?: string;
        /**
         * The JSON Web Token (JWT) to use for authentication.
         * Use with LoginType.CUSTOM
         * See: https://firebase.google.com/docs/auth/server
         */
        token?: string;
        /**
         * A function that returns a promise with the  JSON Web Token (JWT) to use for authentication.
         * Use with LoginType.CUSTOM
         * See: https://firebase.google.com/docs/auth/server
         */
        tokenProviderFn?: () => Promise<String>;
        /**
         * For Facebook auth. Default: ["public_profile", "email"]
         */
        scope?: string[];
    }

    /**
     * The returned object from the login function.
     */
    export interface User {
        uid: string;
        email?: string;
        name?: string;
        provider?: LoginType;
        expiresAtUnixEpochSeconds?: number;
        profileImageURL?: string;
        token?: string;
    }

    /**
     * The returned object from the push function.
     */
    export interface PushResult {
        key: string;
    }

    /**
     * The returned object from the addEventListener functions.
     */
    export interface AddEventListenerResult {
        path: string;
        listeners: Array<any>
    }

    /**
     * The returned object from the createUser function.
     */
    export interface CreateUserResult {
        key: string;
    }

    /**
     * The options object passed into the createUser function.
     */
    export interface CreateUserOptions {
        email: string;
        password: string;
    }

    /**
     * The options object passed into the resetPassword function.
     */
    export interface ResetPasswordOptions {
        email: string;
    }

    /**
     * The returned object in the callback handlers
     * of the addChildEventListener and addValueEventListener functions.
     */
    export interface FBData {
        type: string;
        key: string;
        value: any;
    }

    /**
     * The options object passed into the changePassword function.
     */
    export interface ChangePasswordOptions {
        email: string;
        oldPassword: string;
        newPassword: string
    }

    export interface AuthStateData {
        loggedIn?: boolean;
        user?: User;
    }

    export interface AuthStateChangeListener {
        onAuthStateChanged: (data: AuthStateData) => void;
        thisArg?: any;
    }

    export interface RemoteConfigProperty {
        key: string;
        default: any;
    }

    export interface GetRemoteConfigOptions {
        /**
         * Fetch new results from the server more often.
         * Default false.
         */
        developerMode?: boolean;
        /**
         * The number of seconds before retrieving fresh state from the server.
         * Default 12 hours.
         */
        cacheExpirationSeconds?: number;
        /**
         * The configuration properties to retrieve for your app. Specify as:
         *  properties: [{
         *    key: "holiday_promo_enabled",
         *    default: false
         *  }, ..]
         */
        keys: Array<string>;
        properties: Array<RemoteConfigProperty>
    }

    /**
     * The returned object from the getRemoteConfig function.
     */
    export interface GetRemoteConfigResult {
        /**
         * The date the data was last refreshed from the server.
         * Should honor the 'cacheExpirationSeconds' you passed in previously.
         */
        lastFetch: Date;
        /**
         * The result may be throttled when retrieved from the server.
         * Even when the cache has expired. And it's just FYI really.
         */
        throttled: boolean;
        /**
         * A JS Object with properties and values.
         * If you previously requested keys ["foo", "is_enabled"] then this will be like:
         *   properties: {
       *     foo: "bar",
       *     is_enabled: true
       *   }
         */
        properties: Object;
    }

    /**
     * The returned object in the callback handler of the addOnMessageReceivedCallback function.
     *
     * Note that any custom data you send from your server will be available as
     * key/value properties on the Message object.
     */
    export interface Message {
        /**
         * Indicated whether or not the notification was received while the app was in the foreground.
         */
        foreground: boolean;
        /**
         * The main text shown in the notificiation.
         * Not available on Android when the notification was received in the background.
         */
        body?: string;
        /**
         * Optional title, shown above the body in the notification.
         * Not available on Android when the notification was received in the background.
         */
        title?: string;
        /**
         * iOS badge count, as sent from the server.
         */
        badge?: number;
    }

    /**
     * Use either the 'localFile' or 'localFullPath' param to upload a file.
     */
    export interface UploadFileOptions {
        /**
         * If you didn't pass 'storageBucket' during init() you will need to do it now.
         * Takes the form of 'gs://n-plugin-test.appspot.com' and can be found in the Firebase console.
         */
        bucket?: string;
        /**
         * The full path of the file in your Firebase storage (folders will be created)
         * Example: 'uploads/images/telerik-logo-uploaded.png'
         */
        remoteFullPath: string;
        /**
         * Option 1: a file-system module File object
         * Example:
         *   var fs = require("file-system");
         *   fs.File.fromPath("path-to-the-file")
         */
        localFile?: any;
        /**
         * Option 2: a full file path (ignored if 'localFile' is set)
         */
        localFullPath?: string;

        /**
         * You can get updates during upload by passing a function, example:
         *
         * onProgress: function(status) {
         *   console.log("Uploaded fraction: " + status.fractionCompleted);
         *   console.log("Percentage complete: " + status.percentageCompleted);
         * }
         */
        onProgress?: Function;
    }

    export interface UploadFileResult {
    }

    export interface DownloadFileOptions {
        /**
         * If you didn't pass 'storageBucket' during init() you will need to do it now.
         * Takes the form of 'gs://n-plugin-test.appspot.com' and can be found in the Firebase console.
         */
        bucket?: string;
        /**
         * The full path of an existing file in your Firebase storage
         * Example: 'uploads/images/telerik-logo-uploaded.png'
         */
        remoteFullPath: string;
        /**
         * Option 1: a file-system module File object
         * Example:
         *   var fs = require("file-system");
         *   fs.File.fromPath("path-to-the-file")
         */
        localFile?: any;
        /**
         * Option 2: a full file path (ignored if 'localFile' is set)
         */
        localFullPath?: string
    }

    export interface GetDownloadUrlOptions {
        /**
         * If you didn't pass 'storageBucket' during init() you will need to do it now.
         * Takes the form of 'gs://n-plugin-test.appspot.com' and can be found in the Firebase console.
         */
        bucket?: string;
        /**
         * The full path of an existing file in your Firebase storage
         * Example: 'uploads/images/telerik-logo-uploaded.png'
         */
        remoteFullPath: string;
    }

    export interface DeleteFileOptions {
        /**
         * If you didn't pass 'storageBucket' during init() you will need to do it now.
         * Takes the form of 'gs://n-plugin-test.appspot.com' and can be found in the Firebase console.
         */
        bucket?: string;
        /**
         * The full path of an existing file in your Firebase storage
         * Example: 'uploads/images/telerik-logo-uploaded.png'
         */
        remoteFullPath: string;
    }

    export interface SendCrashLogOptions {
        /**
         * Any custom logging you want to send to Firebase.
         */
        log: string;
    }

    // Analytics, nicely grouped in its own module
    namespace analytics {
        export interface LogEventParameter {
            key: string;
            value: string;
        }
        export interface LogEventOptions {
            /**
             * The name of the event. You can use any name, but it's recommended to use one of
             * the predefined constants. These values are the same for both iOS and Android, so
             * for the complete list see https://firebase.google.com/docs/reference/android/com/google/firebase/analytics/FirebaseAnalytics.Event.html
             */
            key: string;
            /**
             * Each (predefined) event has its own set of optional parameters, see
             * https://firebase.google.com/docs/reference/android/com/google/firebase/analytics/FirebaseAnalytics.Param
             * Example:
             *
             *  parameters: [{
             *    key: "item_name",
             *    value: "abc"
             *  }, ..]
             */
            parameters?: Array<LogEventParameter>
        }

        export interface SetUserPropertyOptions {
            key: string;
            value: string;
        }

        function logEvent(options: LogEventOptions): Promise<any>;
        function setUserProperty(options: SetUserPropertyOptions): Promise<any>;
    }

    interface Firebase {
        init(options: InitOptions): Promise<any>;

        // Database
        push(path: string, value: any): Promise<PushResult>;
        setValue(path: string, value: any): Promise<any>;
        update(path: string, value: any): Promise<any>;
        remove(path: string): Promise<any>;
        query(onValueEvent: (data: FBData) => void, path: string, options: QueryOptions): Promise<any>;
        addChildEventListener(onChildEvent: (data: FBData) => void, path: string): Promise<AddEventListenerResult>;
        addValueEventListener(onValueEvent: (data: FBData) => void, path: string): Promise<AddEventListenerResult>;
        removeEventListeners(listeners: Array<any>, path: string): Promise<any>;
        /**
         * Tells the client to keep its local cache in sync with the server automatically.
         */
        keepInSync(path: string, switchOn: boolean): Promise<any>;

        // Auth
        login(options: LoginOptions): Promise<User>;
        getAuthToken(option: GetAuthTokenOptions): Promise<string>;
        logout(): Promise<any>;
        createUser(options: CreateUserOptions): Promise<CreateUserResult>;
        deleteUser(): Promise<any>;
        resetPassword(options: ResetPasswordOptions): Promise<any>;
        changePassword(options: ChangePasswordOptions): Promise<any>;
        addAuthStateListener(listener: AuthStateChangeListener): boolean;
        removeAuthStateListener(listener: AuthStateChangeListener): boolean;
        hasAuthStateListener(listener: AuthStateChangeListener): boolean;
        getCurrentUser(): Promise<User>;

        // FCM
        addOnMessageReceivedCallback(onMessageReceived: (data: Message) => void): Promise<any>;
        addOnPushTokenReceivedCallback(onPushTokenReceived: (data: string) => void): Promise<any>;

        // remote config
        getRemoteConfig(options: GetRemoteConfigOptions): Promise<GetRemoteConfigResult>;

        // storage
        uploadFile(options: UploadFileOptions): Promise<UploadFileResult>;
        downloadFile(options: DownloadFileOptions): Promise<any>;
        getDownloadUrl(options: GetDownloadUrlOptions): Promise<string>;
        deleteFile(options: DeleteFileOptions): Promise<any>;

        // crash logging
        // sendCrashLog(options: SendCrashLogOptions): Promise<any>;
    }
}

export = firebase;